/*
 * (c) 2002 Ramil Koussiapkoulov <roma@mail.rb.ru> [ViDiSys], www.vidisys.de
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
//#include <sys/io.h>
#include <asm/io.h>

#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>

#include <sys/mman.h>
#include "saa7134-reg.h"

#define TRUE  1
#define FALSE 0

//
// ADVENTURE :
// nach Schreiben auf I2C-Bus muss man (unbekannt lange) warten (Status Bit's sind ok)
// sonst gibt's bei darauffolgendem Read einen Adressierungs-Fehler !!
// da hab' ich beim micronas ein falsches Bit verwendet !! (nicht repariert)
//

static char *str_i2c_status[] = {
	"IDLE", "DONE_STOP", "BUSY", "TO_SCL", "TO_ARB", "DONE_WRITE",
	"DONE_READ", "DONE_WRITE_TO", "DONE_READ_TO", "NO_DEVICE",
	"NO_ACKN", "BUS_ERR", "ARB_LOST", "SEQ_ERR", "ST_ERR", "SW_ERR"
};

enum i2c_status {
	IDLE          = 0,  // no I2C command pending
	DONE_STOP     = 1,  // I2C command done and STOP executed
	BUSY          = 2,  // executing I2C command
	TO_SCL        = 3,  // executing I2C command, time out on clock stretching
	TO_ARB        = 4,  // time out on arbitration trial, still trying
	DONE_WRITE    = 5,  // I2C command done and awaiting next write command
	DONE_READ     = 6,  // I2C command done and awaiting next read command
	DONE_WRITE_TO = 7,  // see 5, and time out on status echo
	DONE_READ_TO  = 8,  // see 6, and time out on status echo
	NO_DEVICE     = 9,  // no acknowledge on device slave address
	NO_ACKN       = 10, // no acknowledge after data byte transfer
	BUS_ERR       = 11, // bus error
	ARB_LOST      = 12, // arbitration lost during transfer
	SEQ_ERR       = 13, // erroneous programming sequence
	ST_ERR        = 14, // wrong status echoing
	SW_ERR        = 15  // software error
};

static char *str_i2c_attr[] = {
	"NOP", "STOP", "CONTINUE", "START"
};

enum i2c_attr {
	NOP           = 0,  // no operation on I2C bus
	STOP          = 1,  // stop condition, no associated byte transfer
	CONTINUE      = 2,  // continue with byte transfer
	START         = 3   // start condition with byte transfer
};

volatile int *SAA_Reg;

void set_addr(volatile int *SAA_ADDR)
{
  SAA_Reg = SAA_ADDR;
}

unsigned char saa_readb(long reg)
{
  return *((unsigned char *)((unsigned char *)SAA_Reg + reg));
}

void saa_writeb(long reg, unsigned char value)
{
  *((unsigned char *)((unsigned char *)SAA_Reg + reg)) = value;
}

long saa_readl(long reg)
{
  return *((unsigned long *)((unsigned char *)SAA_Reg + reg));
}

void saa_writel(long reg, long value)
{
  *((unsigned long *)((unsigned char *)SAA_Reg + reg)) = value;
}

void saa_andorb(long reg, unsigned char mask, unsigned char value)
{
  saa_writeb(reg, (saa_readb(reg) & ~(mask)) | ((value) & (mask)));
}

static inline enum i2c_status i2c_get_status()
{
	enum i2c_status status;
	
	status = (enum i2c_status)(saa_readb(SAA7134_I2C_ATTR_STATUS) & 0x0f);
	//printf( " i2c stat <= %s\n",
	//str_i2c_status[status]);
	return status;
}

static inline void i2c_set_status(enum i2c_status status)
{
	//printf( "i2c stat => %s\n",
	//str_i2c_status[status]);
	saa_andorb(SAA7134_I2C_ATTR_STATUS,0x0f,status);
}

static inline void i2c_set_attr(enum i2c_attr attr)
{
	//printf( "i2c attr => %s\n",
	//str_i2c_attr[attr]);
	saa_andorb(SAA7134_I2C_ATTR_STATUS,0xc0,attr << 6);
}

static inline int i2c_is_error(enum i2c_status status)
{
	switch (status) {
	case NO_DEVICE:
	case NO_ACKN:
	case BUS_ERR:
	case ARB_LOST:
	case SEQ_ERR:
	case ST_ERR:
		return TRUE;
	default:
		return FALSE;
	}
}

static inline int i2c_is_idle(enum i2c_status status)
{
	switch (status) {
	case IDLE:
	case DONE_STOP:
		return TRUE;
	default:
		return FALSE;
	}
}

static inline int i2c_is_busy(enum i2c_status status)
{
	switch (status) {
	case BUSY:
		return TRUE;
	default:
		return FALSE;
	}
}

static int i2c_is_busy_wait()
{
	enum i2c_status status;
	int count;

	for (count = 0; count < I2C_WAIT_RETRY; count++) {
		status = i2c_get_status();
		if (!i2c_is_busy(status))
			break;
		else
			usleep(I2C_WAIT_DELAY);
	}
	if (I2C_WAIT_RETRY == count)
		return FALSE;
	return TRUE;
}

static int i2c_reset()
{
	enum i2c_status status;
	int count;

	//printf("i2c reset\n");
	status = i2c_get_status();
	if (!i2c_is_error(status))
		return TRUE;
	i2c_set_status(status);

	for (count = 0; count < I2C_WAIT_RETRY; count++) {
		status = i2c_get_status();
		if (!i2c_is_error(status))
			break;
		usleep(I2C_WAIT_DELAY);
	}
	if (I2C_WAIT_RETRY == count)
		return FALSE;

	if (!i2c_is_idle(status))
		return FALSE;
	
	i2c_set_attr(NOP);
	return TRUE;
}

static inline int i2c_send_byte(enum i2c_attr attr,
				unsigned char data)
{
	enum i2c_status status;
	unsigned long dword;
/*
	i2c_set_attr(attr);
	saa_writeb(SAA7134_I2C_DATA, data);
*/
  // have to write both attr + data in one 32bit unsigned short
	dword  = saa_readl(SAA7134_I2C_ATTR_STATUS);
	dword &= 0x0f;
	dword |= (attr << 6);
	dword |= ((unsigned long)data << 8);
	dword |= 0x00 << 16; /* 0x0 - 100 Khz, 0x40 - 400 kHz */
	dword |= 0xf0 << 24;
	saa_writel(SAA7134_I2C_ATTR_STATUS, dword);

	//printf( "i2c data => 0x%x\n",data);
	
	if (!i2c_is_busy_wait())
		return -1;
	status = i2c_get_status();
	if (i2c_is_error(status))
		return -1;
	return 0;
}

static inline int i2c_recv_byte()
{
	enum i2c_status status;
	unsigned char data;
	
	i2c_set_attr(CONTINUE);
	if (!i2c_is_busy_wait())
		return -1;
	status = i2c_get_status();
	if (i2c_is_error(status))
		return -1;
	data = saa_readb(SAA7134_I2C_DATA);
	//printf( "i2c data <= 0x%x\n",data);
	return data;
}

int WriteI2C(unsigned char addr, unsigned char subaddr, unsigned char val)
{
	enum i2c_status status;
	int rc;
  status = i2c_get_status();
	if (!i2c_is_idle(status))
		if (!i2c_reset())
			return -1;
	rc = i2c_send_byte(START,addr);
	if (rc < 0)
		 goto err;
	rc = i2c_send_byte(CONTINUE,subaddr);
	if (rc < 0)
		goto err;
	rc = i2c_send_byte(CONTINUE,val);
	if (rc < 0)
		goto err;
	i2c_set_attr(STOP);
	rc = -1;
	if (!i2c_is_busy_wait())
		goto err;
  	status = i2c_get_status();
	if (i2c_is_error(status))
		goto err;
  usleep(20);
	return val;
 err:
	status = i2c_get_status();
	//printf(" ERROR: %s\n",str_i2c_status[status]);
  return rc;
}

int WriteEmpressCommand(unsigned char addr, unsigned char subaddr)
{
	enum i2c_status status;
	int rc;
  status = i2c_get_status();
	if (!i2c_is_idle(status))
		if (!i2c_reset())
			return -1;
	rc = i2c_send_byte(START,addr);
	if (rc < 0)
		 goto err;
	rc = i2c_send_byte(CONTINUE,subaddr);
	if (rc < 0)
		goto err;
	i2c_set_attr(STOP);
	rc = -1;
	if (!i2c_is_busy_wait())
		goto err;
  	status = i2c_get_status();
	if (i2c_is_error(status))
		goto err;
	return 0;
 err:
	status = i2c_get_status();
	//printf(" ERROR: %s\n",str_i2c_status[status]);
  return rc;
}

int WriteI2C_US(unsigned char addr, unsigned char subaddr, unsigned short w_val)
{
	enum i2c_status status;
	int rc;
  status = i2c_get_status();
	if (!i2c_is_idle(status))
		if (!i2c_reset())
			return -1;
	rc = i2c_send_byte(START,addr);
	if (rc < 0)
		 goto err;
	rc = i2c_send_byte(CONTINUE,subaddr);
	if (rc < 0)
		goto err;
	rc = i2c_send_byte(CONTINUE, ((w_val>>8)&0xFF));
	if (rc < 0)
		goto err;
	rc = i2c_send_byte(CONTINUE, (w_val&0xFF));
	if (rc < 0)
		goto err;
	i2c_set_attr(STOP);
	rc = -1;
	if (!i2c_is_busy_wait())
		goto err;
  	status = i2c_get_status();
	if (i2c_is_error(status))
		goto err;
  usleep(20);
	return w_val;
 err:
	status = i2c_get_status();
	//printf(" ERROR: %s\n",str_i2c_status[status]);
  return rc;
}

unsigned char ReadI2C(unsigned char addr, unsigned char subaddr)
{
  unsigned char val = 0;
	enum i2c_status status;
	int rc;
  status = i2c_get_status();
	if (!i2c_is_idle(status))
		if (!i2c_reset())
			return 0;
	rc = i2c_send_byte(START,addr);
	if (rc < 0)
		 goto err;
	rc = i2c_send_byte(CONTINUE,subaddr);
	if (rc < 0)
		goto err;
	rc = i2c_send_byte(START,addr+1);
	if (rc < 0)
		goto err;
	val = i2c_recv_byte();
	i2c_set_attr(STOP);
	rc = 0;
	if (!i2c_is_busy_wait())
		goto err;
  	status = i2c_get_status();
	if (i2c_is_error(status))
		goto err;
	return val;
 err:
	status = i2c_get_status();
	//printf(" ERROR: %s\n",str_i2c_status[status]);
  return rc;
}

unsigned short ReadI2C_US(unsigned char addr, unsigned char subaddr)
{
  unsigned short w_val = 0;
  unsigned char val = 0;
	enum i2c_status status;
	int rc;
  status = i2c_get_status();
	if (!i2c_is_idle(status))
		if (!i2c_reset())
			return 0;
	rc = i2c_send_byte(START,addr);
	if (rc < 0)
		 goto err;
	rc = i2c_send_byte(CONTINUE,subaddr);
	if (rc < 0)
		goto err;
	rc = i2c_send_byte(START,addr+1);
	if (rc < 0)
		goto err;
	val = i2c_recv_byte();
  w_val = (val <<8);
	w_val |= i2c_recv_byte();
	i2c_set_attr(STOP);
	rc = 0;
	if (!i2c_is_busy_wait())
		goto err;
  	status = i2c_get_status();
	if (i2c_is_error(status))
		goto err;
	return w_val;
 err:
	status = i2c_get_status();
	//printf(" ERROR: %s\n",str_i2c_status[status]);
  return rc;
}
