/*
 * (c) 2002 Ramil Koussiapkoulov <roma@mail.rb.ru> [ViDiSys], www.vidisys.de
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <asm/io.h>

#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>

#include <sys/mman.h>
#include "saa7134-reg.h"
#include "pci.h"

void* mapPhysical(off_t offset, size_t length)
{
  int fd = open("/dev/mem", O_RDWR);
  
  if(fd == -1) {
    perror("open");
    exit(1);
  }
  
  int prot    = PROT_READ | PROT_WRITE;
  int flags   = MAP_SHARED;
  
  //	prot |=  PROT_NOCACHE;
  
  void *mem = mmap
    ( 0,
      length,
      prot,
      flags,
      fd,
      offset
      );
  
  if((void*) -1 == mem) {
     perror("mmap");
     exit(1);
  }

  close(fd);
  
  return mem;
}

static const int MAX_STR = 256;

unsigned char GetByte(unsigned char s)
{
  if ((s >= 0x30) && (s <= 0x39))
    return (s - 0x30);
  else  if ((s >= 0x61) && (s <= 0x66))
    return (s - 0x61)+10;
  else
    return 0;
}

long ReadHEX(char* s)
{
  int pos = 0;
  long val = 0;
  if (pos = strspn(s, "0x") != -1)
  {
    s = s + pos+1;
    int exp = 0;
    for (int i=(strlen(s)-1); i>=0; i--)
    {
      if (exp == 0)
        exp = 1;
      else
        exp *= 16;
      val += GetByte(s[i])*exp;
    }
    return val;
  }
  else
  {
    return atoi(s);
  }
}

// =====================================
void run(void)
{
  int cnt, pos;
  char cmd[MAX_STR], command[MAX_STR];
  bool IsEmpress = false;
  bool IsLong = false;
  char s_addr[MAX_STR], s_wert[MAX_STR];
  int adresse, wert;

  // int runlevel;

  while (1)
  {
    printf(">\n");
    if (fgets(cmd, MAX_STR, stdin) == NULL)
      break;

    cnt = strlen(cmd);
    if (cmd[cnt-1] != '\n')
    {
      printf("Line too long\n");
      do			// Rest der Zeile berlesen
      {
	      if (fgets(cmd, MAX_STR, stdin) == NULL)
	        return;

      	cnt = strlen(cmd);
      } while (cmd[cnt-1] != '\n');
      continue;
    }
    if (sscanf(cmd, "%s %n", command, &pos) != 1)
    {
      printf("Invalid command\n");
      continue;
    }
    if (!strcmp(command, "read")) 
    {				// Argumente ignorieren
      int nrparms = sscanf(cmd+pos, "%s", s_addr);

      if (nrparms < 1)
        printf("Invalid parameter\n");
      else
      {
        if (IsEmpress)
        {
          if (IsLong)
            printf("%x\n",ReadI2C_US(EMPRESS_ADDR, ReadHEX(s_addr)));
          else
            printf("%x\n",ReadI2C(EMPRESS_ADDR, ReadHEX(s_addr)));
        }
        else
        {
          if (IsLong)
            printf("%x\n",saa_readl(ReadHEX(s_addr)));
          else
            printf("%x\n",saa_readb(ReadHEX(s_addr)));
        }
      }
    }
    else if (!strcmp(command, "write"))
    {
      int nrparms = sscanf(cmd+pos, "%s %s", s_addr, s_wert);

      if (nrparms < 2)
        printf("Invalid parameter\n");
      else
      {
        if (IsEmpress)
        {
          if (IsLong)
            WriteI2C_US(EMPRESS_ADDR, ReadHEX(s_addr), ReadHEX(s_wert));
          else
            WriteI2C(EMPRESS_ADDR, ReadHEX(s_addr), ReadHEX(s_wert));
        }
        else
        {
          if (IsLong)
            saa_writel(ReadHEX(s_addr), ReadHEX(s_wert));
          else
            saa_writeb(ReadHEX(s_addr), ReadHEX(s_wert));
        }
      }
    }
    else if (!strcmp(command, "command"))
    {
      int nrparms = sscanf(cmd+pos, "%d", &adresse);

      if (nrparms < 1)
        printf("Invalid parameter\n");
      else
        WriteEmpressCommand(EMPRESS_ADDR, adresse);
    }
    else if (!strcmp(command, "empress")) 
    {
      IsEmpress = true;
    }
    else if (!strcmp(command, "saa7134"))
    {
      IsEmpress = false;
    }
    else if (!strcmp(command, "byte")) 
    {
      IsLong = false;
    }
    else if (!strcmp(command, "long")) 
    {
      IsLong = true;
    }
    else if (!strcmp(command, "exit"))
    {
      break;
    }
    else 
    {
      printf("status:");
      if (IsEmpress)
        printf(" empreess");
      else
        printf(" saa7134");
      if (IsLong)
        printf(" long\n");
      else
        printf(" byte\n");
      printf("usage:\n"
    	     "read [adresse]\n"
             "write [adresse] [wert]\n"
             "command [adresse]\n"
             "empress\n"
             "saa7134\n"
             "byte\n"
             "long\n"
	           "exit\n");
    }
  }
}


int main(int argc, char **argv)
{

  //---------------------------------------------------------------------------
  // Wo steht SAA,VGA ?
  // ist Memory-Mapped !
  //---------------------------------------------------------------------------

  int SAA_Space;
  
  if(scanpci_for_Device(&SAA_Space,SAA_ID))
    printf("SAA_Space = 0x%x \n",SAA_Space);
  else {
    fprintf(stderr," SAA7146 not found !\n");
    exit(1);
  }

  // wenn Hardware ins Memory schreibt physikalische Addressen verwenden : PHYS_....
  // wenn Software von/in Memory schreibt logische Addressen verwenden 

  volatile int *base_addr = (int *) mapPhysical(SAA_Space,4096);      // Memory besorgen
  set_addr(base_addr);
  run();

  exit(0);

}



